/*
	Donut3D
	Version : 1.2
	History
	- 30-Aug-2018 (Ronen Avidor) add error handling in process result when fetching color of each slice
	- 30-Aug-2018 (Ronen Avidor) update styler form and fix label display options
*/
prism.registerWidget("donut3d", {

	name : "donut3d",
	family : "Donut3D",
	title : "Donut 3D",
	iconSmall : "/plugins/Donut3D/donut3d.png",
	styleEditorTemplate: "/plugins/Donut3D/styler.html",
	style: {
		PieType : 'Classic',
		PieHeight: 30,
		ShowCategories: false,
		ShowValue:false,
		ShowPercent:false,
		LabelsEnabled:true
	},
	data : {
		selection : [],
		defaultQueryResult : {},	
		panels : [
			{
				name: 'Category',
				type: "series",
				itemAttributes: ["color"],
				allowedColoringTypes: function() {
					return {
						color: true,
						condition: false,
						range: false
					}
				},
				metadata: {
					types: ['dimensions'],
					maxitems: 1
				},
				visibility: function(w) {
					return w.metadata.panel("Value").items.length < 2
				}
			},
			{
				name: 'Value',
				type: 'visible',	
				itemAttributes: ["color"],
				allowedColoringTypes: function() {
					return  {
						color: true,
						condition: false,
						range: false
					};
				},
				metadata: {
					types: ['measures'],
					maxitems: function(w) {
						if (w.metadata.panel("Category").items.length == 0) {
							return -1;
						} else {
							return 1;
						}
					}
				},
				itemAdded: function(widget, item) {
					var colorFormatType = $$get(item, "format.color.type");
					if ("color" === colorFormatType || "color" === colorFormatType) {
						var color = item.format.color.color;
						defined(color) && "transparent" != color && "white" != color && "#fff" != color && "#ffffff" != color || $jaql.resetColor(item)
					}
					"range" === colorFormatType && $jaql.resetColor(item), defined(item, "format.color_bkp") && $jaql.resetColor(item), defined(item, "format.members") && delete item.format.members
				}
			},			
			{
				name: 'filters',
				type: 'filters',
				metadata: {
					types: ['dimensions'],
					maxitems: -1
				}
			}
		],
		
		canColor: function (widget, panel, item) {
			return (panel.name === "Value" && widget.metadata.panel("Category").items.length == 0);
            //return panel.name === "value" ;
        },
		
		allocatePanel: function (widget, metadataItem) {			
			// measure
			if (prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("Value").items.length === 0) {

				return "Value";
			}
			// dimension
			else if (!prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("Category").items.length < 3) {

				return "Category";
			}
		},

		// returns true/ reason why the given item configuration is not/supported by the widget
		isSupported: function (items) {

			return this.rankMetadata(items, null, null) > -1;
		},

		// ranks the compatibility of the given metadata items with the widget
		rankMetadata: function (items, type, subtype) {

			var a = prism.$jaql.analyze(items);

			// require 1 measure, 2 dimension
			if (a.measures.length == 2 && a.dimensions.length == 1) {

				return 0;
			}

			return -1;
		},

		// populates the metadata items to the widget
		populateMetadata: function (widget, items) {

			var a = prism.$jaql.analyze(items);
			
			
			// allocating dimensions
			widget.metadata.panel("Category").push(a.dimensions);
			widget.metadata.panel("Value").push(a.measures);

			// allocating filters
			widget.metadata.panel("filters").push(a.filters);
		},
		
		// builds a jaql query from the given widget, this will also populate the Category dimension with all values and a color box next to it
		buildQuery: function (widget) {
			// building jaql query object from widget metadata 
			var query = { 
				datasource: widget.datasource, 
				format: "json",
				isMaskedResult:true,
				metadata: [] 
			};
			
			
			// Categories items
			widget.metadata.panel("Category").items.forEach(function (item) {

				query.metadata.push(item);
			});

			// series
			widget.metadata.panel("Value").items.forEach(function (item) {

				query.metadata.push(item);
			});

			// filters
			widget.metadata.panel("filters").items.forEach(function (item) {

				var filteritem = $$.object.clone(item, true);
				filteritem.panel = "scope";
				filteritem.filterType = 'widget';
				
				query.metadata.push(filteritem);
			});
				
			
			return query;
			
		},

		//Create Donut3d Data Object and populate all values of the pie
		processResult : function (widget, queryResult) {	
		
			var newResults = {Initial:[], Final:[]};
			// Figure out how the dashboard filter was select, included item, execluded item or all
			if (prism.activeDashboard.filters.$$items.length > 0) {
				var dashboardFilters =  prism.activeDashboard.filters.$$items[0].jaql.all ? null: prism.activeDashboard.filters.$$items[0].jaql.filter.members;
				var AllFilterSelected = prism.activeDashboard.filters.$$items[0].jaql.filter.all? true: false;
				var ExcludedFilterSelected = prism.activeDashboard.filters.$$items[0].jaql.filter.exclude? prism.activeDashboard.filters.$$items[0].jaql.filter.exclude.members: null;
			};
			
			if (!dashboardFilters) {
				Donut3D.PreviousSelection = -1;
				
			}
			
			// Calculate Total for all values
			var totalValues = 0;
			for (i=0; i<queryResult.$$rows.length; i++) {
				totalValues += queryResult.$$rows[i][1].data;
			};
			
			var hasCategories = (widget.metadata.panels[0].items.length > 0);
			
			if (hasCategories) {
				//using categories
				var categoryColors = widget.metadata.panels[0].items[0].format.members;
				
				//	Create the data objects
				for (i=0; i<queryResult.$$rows.length; i++) {
					
					//	Get the data values needed
					var thisText = queryResult.$$rows[i][0].text,
						thisData = ($$get(queryResult, '$$rows.' + i + '.1.data') || 1).toString(),
						thisDataFormatted = ($$get(queryResult, '$$rows.' + i + '.1.text') || "1"),
						thisColor,
						ItemSelected = defined(queryResult.$$rows[i][0].selected) ? queryResult.$$rows[i][0].selected : true
						
						
					try {
						thisColor = categoryColors[queryResult.$$rows[i][0].text].color
					}
					catch(err) {
						thisColor = 'black'
					}						
					
					var thisItem = {label:0, value:0, formattedtext: "",percent:0, color:"",selected:false,WidgetObject:widget,DashboardFiltered:true,ValueLabelText:""}
					thisItem.label = thisText;
					thisItem.value = thisData;
					thisItem.percent = thisData / totalValues;
					thisItem.color = thisColor;
					thisItem.formattedtext = thisDataFormatted;
					thisItem.id = i;
					
					// Set a flag to determine if this member is selected within the dashboard filters
					if (prism.activeDashboard.filters.$$items.length > 0)  {
						thisItem.DashboardFiltered = ExcludedFilterSelected? !(ExcludedFilterSelected.includes(thisText)) : ( AllFilterSelected? true : (dashboardFilters ? dashboardFilters.includes(thisText): true));
					};

					thisItem.ValueLabelText = (widget.metadata.panels[1].items.length > 0) ? widget.metadata.panels[1].items[0].jaql.title: "";
					
					var thisItemInitial = {label:0, value:0,formattedtext: "", percent:0,color:"",selected:false,WidgetObject:widget,DashboardFiltered:true};
					thisItemInitial.label = thisText;
					thisItemInitial.value = 1000*Math.random();
					thisItemInitial.color = thisColor;
					thisItemInitial.id = i;
					thisItemInitial.formattedtext = thisDataFormatted;
					// Set a flag to determine if this member is selected within the dashboard filters
					if (prism.activeDashboard.filters.$$items.length > 0) {
						thisItemInitial.DashboardFiltered = ExcludedFilterSelected? !(ExcludedFilterSelected.includes(thisText)) : ( AllFilterSelected? true : (dashboardFilters ? dashboardFilters.includes(thisText): true));
					};
					
					//if (ItemSelected) {
						newResults.Initial.push(thisItemInitial);
						newResults.Final.push(thisItem);
						//newResults.Initial[i] = thisItemInitial;
						//newResults.Final[i] = thisItem;
					//}
					
				}
			} 
		
			return  newResults;
		}
		
	},
	
	beforequery: function(widget, args){
		
		
		//	Need to check only the first item
		var nodeDim = args.query.metadata[0];

		//	Look for any dashboard filters
		var dashboardFilters = prism.activeDashboard.filters.$$items;

		//	Loop through the metadata
		for (var i=args.query.metadata.length-1; i>0; i--){

			//	Does this item match the node's dimension AND is a dashboard filter?
			var isMatch = (args.query.metadata[i].panel == "scope") && (args.query.metadata[i].jaql.dim == nodeDim.jaql.dim),
				//isNotWidgetFilter = (args.query.metadata[i].filterType !== "widget"),
				isWidgetFilter = (args.query.metadata[i].filterType == "widget"),
				isDashboardFilter = $.grep(dashboardFilters, function(w){
					return (w.jaql.dim == args.query.metadata[i].jaql.dim) && (typeof w.jaql.filter.all == "undefined");
				}).length > 0;
			//if (isMatch && isNotWidgetFilter && isDashboardFilter) {
			if (isMatch) {
				
				//	Create a new copy of the filter object
				var matchFilter = $$.object.clone(args.query.metadata[i], true);

				//	Remove the panel attribute
				delete matchFilter.panel

				//	Include with the node's dimension metadata
				//nodeDim.jaql.in = {selected: matchFilter}
				if (isWidgetFilter && defined(matchFilter.jaql.filter)) {
					nodeDim.jaql.filter = matchFilter.jaql.filter;
				}

				//	Remove from the metadata
				args.query.metadata.splice(i,1);
			}
		}		
	},

		
		
	render : function (widget, event) {
		var element = $(event.element);
		var innerRadius = 0;
		
		element.empty();

		var svgHeight = $(element).height() - 60;
		var svgWidth = $(element).width() - 40;
		var svgID = widget.oid ? 'donut3d-' + widget.oid : 'donut3d';
		
		var svg = d3.select("body").append("svg").attr("width",'97%').attr("height",'97%');
		
		svg.append("g").attr("id","'+ svgID +'");
		element.append('<svg width="97%" height="97%"><g id="' + svgID + '"></g></svg>');

		// Set the pie type based on the widget design selection
		switch (widget.style.PieType){
			case 'Classic':
				innerRadius = 0;
				break;
			case 'Donut':
				innerRadius = 0.4;
				break;
			case 'Ring':
				innerRadius = 0.8;
				break;
			default:
				innerRadius = 0;
		};
		
		
		if (widget.queryResult.Initial){
			Donut3D.draw(svgID, widget.queryResult.Initial, svgWidth/2, svgHeight/2, svgWidth/4, svgHeight/4, widget.style.PieHeight, innerRadius);
			
			 setTimeout(function(){ 

							Donut3D.transition(svgID, widget.queryResult.Final, svgWidth/4, svgHeight/4, widget.style.PieHeight, innerRadius,widget.style);
							
						}, 20
			);
		}
	},


	destroy : function (s, e) {}
});